/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.formatter;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.Collection;
import java.util.Date;
import java.util.Deque;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map.Entry;
import java.util.logging.Formatter;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.SimpleFormatter;

import org.ow2.petals.commons.log.LogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.slf4j.MDC;

/**
 * <p>This formatter is aimed at using first {@link LogRecord} parameter as "just in time evaluated" extra log information.</p>
 * 
 * TODO use {@link MDC} instead!
 * <p>
 * It also prefix the log message with the container name if the information is available (see
 * {@link PetalsExecutionContext}).
 * </p>
 * <p>Note that localization is still possible (parameters are shifted left if necessary).</p>
 * <p>
 * In other words :
 * </p>
 * <ul>
 * <li>if first parameter is an instance of {@link LogData}, this data will be appended in the log message,</li>
 * <li>otherwise, first parameter is kept for localization duty, as usual.</li>
 * </ul>
 * <p>
 * This class is partly inspired by {@link SimpleFormatter}. It is not possible to directly use it because our format
 * has extra parameters.
 * </p>
 * 
 * @author mjambert
 * @author vnoel
 * 
 */
public class LogDataFormatter extends Formatter {

    private static final String PREFIX_PROPERTY_NAME = LogDataFormatter.class.getName();

    private static final String FORMAT_PROPERTY_NAME = PREFIX_PROPERTY_NAME + ".format";

    private static final String APPENDER_PROPERTY_NAME = PREFIX_PROPERTY_NAME + ".appender";

    private static final String STARTING_DELIMITER_PROPERTY_NAME = PREFIX_PROPERTY_NAME + ".starting-delimiter";

    private static final String ENDING_DELIMITER_PROPERTY_NAME = PREFIX_PROPERTY_NAME + ".ending-delimiter";

    private static final String DEFAULT_FORMAT = "%5$s %1$tY/%<tm/%<td %<tH:%<tM:%<tS,%<tL GMT%<tz %4$s [%3$s] : %6$s%7$s%n";

    private static final String DEFAULT_STARTING_DELIMITER = "";

    private static final String DEFAULT_ENDING_DELIMITER = "";

    /**
     * format string for printing the log record
     */
    private final String format;

    /**
     * TODO not sure there is really a reason to have this dataAppenders as a {@link Collection}...
     */
    private final Deque<LogDataAppender> dataAppenders;

    private final String startingDelimiter;

    private final String endingDelimiter;

    private final Date dat = new Date();

    public LogDataFormatter() {
        this(getFormat(), getStartingDelimiter(), getEndpoingDelimiter(), getLogDataAppender());
    }

    private static String getFormat() {
        String format = LogManager.getLogManager().getProperty(FORMAT_PROPERTY_NAME);
        if (format != null) {
            try {
                // validate the user-defined format string
                String.format(format, new Date(), "", "", "", "", "");
            } catch (IllegalArgumentException e) {
                // illegal syntax; fall back to the default format
                format = DEFAULT_FORMAT;
            }
        } else {
            format = DEFAULT_FORMAT;
        }
        return format;
    }

    private static LogDataAppender getLogDataAppender() {
        final String appender = LogManager.getLogManager().getProperty(APPENDER_PROPERTY_NAME);
        LogDataAppender logDataAppender;
        if (appender != null && !appender.trim().isEmpty()) {
            try {
                final Class<?> appenderClz = ClassLoader.getSystemClassLoader().loadClass(appender);
                final Object appenderObj = appenderClz.newInstance();
                if (appenderObj instanceof LogDataAppender) {
                    logDataAppender = (LogDataAppender) appenderObj;
                } else {
                    logDataAppender = new BasicLogDataAppender();
                }
            } catch (final ClassNotFoundException | InstantiationException | IllegalAccessException e) {
                logDataAppender = new BasicLogDataAppender();
            }
        } else {
            logDataAppender = new BasicLogDataAppender();
        }
        return logDataAppender;
    }

    private static String getStartingDelimiter() {
        final String startingDelimiter = LogManager.getLogManager().getProperty(STARTING_DELIMITER_PROPERTY_NAME);
        if (startingDelimiter != null && !startingDelimiter.trim().isEmpty()) {
            return startingDelimiter;
        } else {
            return DEFAULT_STARTING_DELIMITER;
        }
    }

    private static String getEndpoingDelimiter() {
        final String endingDelimiter = LogManager.getLogManager().getProperty(ENDING_DELIMITER_PROPERTY_NAME);
        if (endingDelimiter != null && !endingDelimiter.trim().isEmpty()) {
            return endingDelimiter;
        } else {
            return DEFAULT_ENDING_DELIMITER;
        }
    }

    LogDataFormatter(final String startingDelimiter, final String endingDelimiter) {
        this(getFormat(), startingDelimiter, endingDelimiter, getLogDataAppender());
    }

    public LogDataFormatter(final String format, final String startingDelimiter, final String endingDelimiter,
            final LogDataAppender logDataAppender) {
        this.format = format;
        this.startingDelimiter = startingDelimiter;
        this.endingDelimiter = endingDelimiter;
        this.dataAppenders = new LinkedList<>();
        this.dataAppenders.add(logDataAppender);
    }

    public final void addLogDataAppender(LogDataAppender logDataAppender) {
        assert logDataAppender != null;
        dataAppenders.add(logDataAppender);
    }

    final LogDataAppender[] getLogDataAppenders() {
        return dataAppenders.toArray(new LogDataAppender[dataAppenders.size()]);
    }

    /**
     * This is the same as {@link SimpleFormatter#format(LogRecord)} but with an extra parameter just before message for
     * the container name.
     */
    @Override
    public synchronized String format(LogRecord record) {
        dat.setTime(record.getMillis());
        String source;
        if (record.getSourceClassName() != null) {
            source = record.getSourceClassName();
            if (record.getSourceMethodName() != null) {
                source += " " + record.getSourceMethodName();
            }
        } else {
            source = record.getLoggerName();
        }

        final Object[] parameters = record.getParameters();
        final LogData extraParameters = getExtraParameters(parameters);

        if (extraParameters != null) {
            Object[] parametersWithoutLogData = Arrays.copyOfRange(parameters, 1, parameters.length);
            record.setParameters(parametersWithoutLogData);
        }

        final String msg = formatMessage(record);

        final String message;
        if (extraParameters != null && !extraParameters.isEmpty()) {
            record.setParameters(parameters);
            final StringBuilder sb = new StringBuilder(msg).append(" ");
            formatExtraParameters(sb, extraParameters);
            message = sb.toString();
        } else {
            message = msg;
        }

        final String throwable;
        if (record.getThrown() != null) {
            final StringWriter sw = new StringWriter();
            try (final PrintWriter pw = new PrintWriter(sw)) {
                pw.println();
                record.getThrown().printStackTrace(pw);
            }
            throwable = sw.toString();
        } else {
            throwable = "";
        }

        String containerName = PetalsExecutionContext.getContainerName();
        if (containerName == null) {
            containerName = "";
        }

        // note: in SimpleFormatter, getLevel().getLocalizedLevelName() is used, but in our formatter, we prefer to
        // always use the english names (and getLocalizedLevelName du not visible anyway...)
        return String.format(format, dat, source, record.getLoggerName(), record.getLevel().getName(),
                containerName, message, throwable);
    }

    protected static LogData getExtraParameters(final Object[] logParameters) {
        if (logParameters != null && logParameters.length != 0 && logParameters[0] instanceof LogData) {
            return (LogData) logParameters[0];
        } else {
            return null;
        }
    }

    protected void formatExtraParameters(final StringBuilder outBuffer, final LogData extraParameters) {
        if (extraParameters.isEmpty() || dataAppenders.isEmpty()) {
            return;
        }

        outBuffer.append(this.startingDelimiter);

        boolean isFirstAppending = true;
        for (final Entry<String, Object> entry : extraParameters.entrySet()) {
            final Iterator<LogDataAppender> it = dataAppenders.descendingIterator();
            while (it.hasNext()) {
                final LogDataAppender appender = it.next();
                if (!appender.canAppend(extraParameters, entry.getKey(), entry.getValue())) {
                    continue;
                }
                if (isFirstAppending) {
                    isFirstAppending = false;
                } else {
                    outBuffer.append(", ");
                }
                appender.append(outBuffer, extraParameters, entry.getKey(), entry.getValue());
                break;
            }
        }

        outBuffer.append(this.endingDelimiter);
    }
}
