/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.handler;

import java.io.File;
import java.io.IOException;
import java.util.logging.ErrorManager;
import java.util.logging.LogRecord;

import javax.jbi.messaging.MessageExchange;

import org.apache.commons.pool2.impl.GenericObjectPoolConfig;
import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.LogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.commons.log.TraceCode;

import com.ebmwebsourcing.easycommons.io.FileSystemHelper;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * <p>
 * Abstract class for contextual file handler based on {@link PetalsFileHandler} and dumping message content.
 * </p>
 */
public abstract class AbstractPetalsDumperFileHandler extends PetalsFileHandler {

    public static final String DUMP_FILE_LOGDATA_ATTR_NAME = "dumpFile";

    protected static final String DUMP_BASEDIR_PROPERTY_NAME = ".dump-basedir";

    protected static final String TRANSFORMERS_POOL_MAX_ACTIVE_PROPERTY_NAME = ".transformersPoolMaxActive";

    protected static final String TRANSFORMERS_POOL_MAX_IDLE_PROPERTY_NAME = ".transformersPoolMaxIdle";

    protected static final String JBI_OUT_MESSAGE = "out";

    protected static final String JBI_IN_MESSAGE = "in";

    private static final String FILENAME_SEPARATOR = "_";

    protected final SourceHelpers sourceHelpers;

    /**
     * The root directory (absolute) of sub directory of dumps.
     * 
     * If the value of the configuration property {@value #DUMP_BASEDIR_PROPERTY_NAME} is not an absolute path, the dump
     * dir is defined as relative to {@link #flowsSubdir} .
     */
    protected final File dumpBasedir;

    private final String dumpFileExtension;

    protected AbstractPetalsDumperFileHandler(final String dumpFileExtension)
            throws PropertiesException {
        super();

        final int transformerPoolMaxActive = getIntProperty(TRANSFORMERS_POOL_MAX_ACTIVE_PROPERTY_NAME,
                GenericObjectPoolConfig.DEFAULT_MAX_TOTAL);

        final int transformerPoolMaxIdle = getIntProperty(TRANSFORMERS_POOL_MAX_IDLE_PROPERTY_NAME,
                GenericObjectPoolConfig.DEFAULT_MAX_IDLE);

        this.sourceHelpers = new SourceHelpers(transformerPoolMaxActive, transformerPoolMaxIdle);

        // Initialize the dump base dir
        final String dumpBaseDirValue = getStringProperty(DUMP_BASEDIR_PROPERTY_NAME,
                this.flowsSubdir.getAbsolutePath());

        final File dumpBasedirPath = new File(dumpBaseDirValue);
        if (dumpBasedirPath.isAbsolute()) {
            this.dumpBasedir = dumpBasedirPath;
        } else {
            this.dumpBasedir = new File(this.flowsSubdir, dumpBaseDirValue);
        }

        this.dumpFileExtension = dumpFileExtension;
    }

    @Override
    public synchronized void publish(final LogRecord record) {

        int levelValue = getLevel().intValue();
        if (record.getLevel().intValue() >= levelValue && levelValue != offValue) {

            try {
                this.publishLogRecord(record);
            } catch (final Exception e) {
                reportError("Error while dumping message exchange", e, ErrorManager.WRITE_FAILURE);
            }

            this.printLog(record);
        }
    }

    private void publishLogRecord(final LogRecord record) throws Exception {

        final Object[] logParameters = record.getParameters();
        if (Level.MONIT.equals(record.getLevel()) && logParameters != null && logParameters.length != 0
                && logParameters[0] != null && logParameters[0] instanceof LogData) {

            final LogData logData = (LogData) logParameters[0];
            final String flowInstanceId = (String) logData.get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME);
            final String flowStepId = (String) logData.get(FlowLogData.FLOW_STEP_ID_PROPERTY_NAME);

            assert flowInstanceId != null : "A MONIT log event MUST always have an attribute 'flowInstanceId' set !!!";
            assert !flowInstanceId.trim().isEmpty() : "A MONIT log event MUST always have a non-empty attribute 'flowInstanceId' !!!";
            assert flowStepId != null : "A MONIT log event MUST always have an attribute 'flowStepId' set !!!";
            assert !flowStepId.trim().isEmpty() : "A MONIT log event MUST always have a non-empty attribute 'flowStepId' !!!";

            final MessageExchange exchange = (MessageExchange) logData.get(FlowLogData.FLOW_EXCHANGE_PROPERTY_NAME);
            if (exchange != null) {

                final TraceCode traceCode = (TraceCode) logData.get(FlowLogData.FLOW_STEP_TRACE_CODE);

                assert traceCode != null : "A MONIT log event MUST always have an attribute 'traceCode' set !!!";

                final File dumpFile = this.getDumpFile(traceCode, flowInstanceId, flowStepId);
                logData.putData(DUMP_FILE_LOGDATA_ATTR_NAME,
                        FileSystemHelper.getRelativePath(this.basedir.getAbsoluteFile(), dumpFile));
                dumpFile.getParentFile().mkdirs();

                this.dumpMessageExchange(exchange, dumpFile);
            }
        }
    }

    /**
     * Dump the message exchange in a file
     * 
     * @param exchange
     *            The message exchange to dump. Not {@code null}.
     * @param toFile
     *            The file in which the message exchange us dumped. Not {@code null}.
     */
    protected abstract void dumpMessageExchange(final MessageExchange exchange, final File toFile) throws Exception;

    /**
     * <p>
     * Generate a {@link File} with an absolute path to dump message into the directory associated to the given flow
     * instance identifier.
     * </p>
     * 
     * @param providedFlowInstanceId
     *            A flow instance identifier. If <code>null</code>, the flow instance identifier is retrieved from the
     *            current context.
     * 
     * @return
     * @throws IOException
     */
    private final File getDumpFile(final TraceCode traceCode, final String providedFlowInstanceId,
            final String flowStepId) {

        final String flowInstanceId;
        if (providedFlowInstanceId == null) {
            flowInstanceId = PetalsExecutionContext.getFlowInstanceId();
        } else {
            flowInstanceId = providedFlowInstanceId;
        }
        final String dumpFileName = this.getContextualDumpFileName(traceCode, flowStepId);
        return new File(new File(this.dumpBasedir, flowInstanceId), dumpFileName);
    }

    /**
     * Compute the dump file name:
     * <code>&lt;flow-step-id&gt;_&lt;trace-code&gt;.xml</code> where:
     * <ul>
     * <li><code>flow-step-id</code> is the flow step identifier, or
     * <code>unknown</code> if the flow step identifier exists,</li>
     * <li><code>trace-code</code> is a code defining the beginning or ending of
     * the current flow step.</li>
     * </ul>
     * 
     * @param traceCode
     *            A code defining the beginning or ending of the current flow
     *            step
     * @return The dump file name to save the content of the message
     */
    private final String getContextualDumpFileName(final TraceCode traceCode,
            final String flowStepId) {
        return (flowStepId == null ? "unknown" : flowStepId) + FILENAME_SEPARATOR + traceCode
                + this.dumpFileExtension;
    }
}
