/**
 * Copyright (c) 2017-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.handler;


import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Set;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;

import com.ebmwebsourcing.easycommons.properties.PropertiesException;

/**
 * <p>
 * A contextual file handler based on {@link PetalsFileHandler} and dumping message exchange content.
 * </p>
 */
public class PetalsExchangeDumperFileHandler extends AbstractPetalsDumperFileHandler {

    private static final String DUMP_FILE_EXTENSION = ".log";

    private static final String NORMALIZED_MESSAGE_HEADER = "=====  %s =====%n";

    public PetalsExchangeDumperFileHandler() throws PropertiesException {
        super(DUMP_FILE_EXTENSION);
    }

    @Override
    protected void dumpMessageExchange(final MessageExchange exchange, final File toFile) throws Exception {

        assert exchange != null;
        assert toFile != null;

        try (final FileOutputStream fos = new FileOutputStream(toFile); final PrintWriter pw = new PrintWriter(fos);) {

            pw.write(String.format("MEP exchange id: %s%n", exchange.getExchangeId()));
            pw.write(String.format("MEP exchange: %s%n", exchange.getPattern()));
            pw.write(String.format("MEP status: %s%n", exchange.getStatus()));
            pw.write(String.format("Interface: %s%n", exchange.getInterfaceName()));
            pw.write(String.format("Service: %s%n", exchange.getService()));
            pw.write(String.format("Endpoint: %s%n", exchange.getEndpoint()));
            pw.write(String.format("Operation: %s%n", exchange.getOperation()));
            pw.write('\n');

            final Set<String> propertyNames = exchange.getPropertyNames();
            if (propertyNames != null) {
                pw.write("Message exchange properties:\n");
                for (final String propertyName : propertyNames) {
                    pw.write(String.format("\t%s : %s%n", propertyName, exchange.getProperty(propertyName)));
                }
            } else {
                pw.write("Message exchange properties: no property defined\n");
            }
            pw.write('\n');

            final NormalizedMessage inNm = exchange.getMessage(JBI_IN_MESSAGE);
            if (inNm != null) {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "IN message"));
                dumpNormalizedMessage(inNm, pw);
            } else {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "No IN message"));
            }
            pw.write('\n');

            final NormalizedMessage outNm = exchange.getMessage(JBI_OUT_MESSAGE);
            if (outNm != null) {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "OUT message"));
                dumpNormalizedMessage(outNm, pw);
            } else {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "No OUT message"));
            }
            pw.write('\n');

            final NormalizedMessage faultNm = exchange.getFault();
            if (faultNm != null) {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "Fault message"));
                dumpNormalizedMessage(faultNm, pw);
            } else {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "No fault message"));
            }
            pw.write('\n');

            final Exception error = exchange.getError();
            if (error != null) {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "Error message"));
                error.printStackTrace(pw);
            } else {
                pw.write(String.format(NORMALIZED_MESSAGE_HEADER, "No error message"));
            }
            pw.write('\n');
        }
    }

    private void dumpNormalizedMessage(final NormalizedMessage nm, final PrintWriter pw) throws Exception {

        this.dumpNmProperties(nm, pw);
        pw.write('\n');
        this.dumpNmContent(nm, pw);
        pw.write('\n');
        this.dumpNmAttachments(nm, pw);
    }

    private void dumpNmProperties(final NormalizedMessage nm, final PrintWriter pw) {

        final Set<String> propertyNames = nm.getPropertyNames();
        if (propertyNames != null) {
            if (propertyNames.isEmpty()) {
                pw.write("Normalized message properties: no property defined\n");
            } else {
                pw.write("Normalized message properties:\n");
                for (final String propertyName : propertyNames) {
                    pw.write(String.format("\t%s : %s%n", propertyName, nm.getProperty(propertyName)));
                }
            }
        } else {
            pw.write("Normalized message properties: no property defined\n");
        }
    }

    private void dumpNmContent(final NormalizedMessage nm, final PrintWriter pw) throws Exception {

        final Source source = nm.getContent();
        if (source != null) {
            try {
                pw.write("Normalized message content:\n");
                this.sourceHelpers.toFile(source, pw);
                pw.write('\n');
            } catch (final Exception e) {
                pw.write("Error dumping normalized message:");
                e.printStackTrace(pw);
                throw e;
            }
        } else {
            pw.write("Normalized message content: EMPTY\n");
        }
    }

    private void dumpNmAttachments(final NormalizedMessage nm, final PrintWriter pw) {

        final Set<String> attachmentNames = nm.getAttachmentNames();
        if (attachmentNames != null) {
            if (attachmentNames.isEmpty()) {
                pw.write("Normalized message attachements: no attachements\n");
            } else {
                pw.write("Normalized message attachements:\n");
                final String attachmentContent = "the attachment will be dumped in a next version of the log handler";
                for (final String attachmentName : attachmentNames) {
                    final DataHandler dh = nm.getAttachment(attachmentName);
                    try {
                        if (dh != null && dh.getInputStream() != null) {
                            final int attachmentSize = dh.getInputStream().available();
                            pw.write(String.format("\t%s (size: %d bytes) : %s%n", attachmentName, attachmentSize,
                                    attachmentContent));
                        } else {
                            pw.write(String.format("\t%s (size: 'not available') : %s%n", attachmentName,
                                    attachmentContent));
                        }
                    } catch (final IOException e) {
                        pw.write(String.format("\t%s (size: 'an error occurs when computing size') : %s%n",
                                attachmentName, attachmentContent));
                        e.printStackTrace(pw);
                    }
                }
            }
        } else {
            pw.write("Normalized message attachements: no attachements\n");
        }
    }
}
