/**
 * Copyright (c) 2011-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.handler;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.logging.ErrorManager;
import java.util.logging.Handler;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;
import java.util.logging.StreamHandler;

import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.commons.log.LogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;

import com.ebmwebsourcing.easycommons.properties.PropertiesException;
import com.ebmwebsourcing.easycommons.properties.PropertiesHelper;

/**
 * <p>
 * The contextual file handler of Petals to use as log handler ({@link Handler})
 * to print monitoring traces with following configuration properties:
 * </p>
 * <ul>
 * <li>the property <code>basedir</code>, default value set to
 * <code>${petals.log.dir}</code>,</li>
 * <li>the property <code>subdir</code>, default value set to
 * <code>flow-monitoring</code>,</li>
 * <li>the property <code>logFilename</code> is set to <code>petals.log</code></li>
 * </ul>
 */
public class PetalsFileHandler extends StreamHandler {

    protected static final int offValue = Level.OFF.intValue();

    protected static final String BASEDIR_PROPERTY_NAME = ".basedir";

    public static final String PETALS_LOG_DIR_PROPERTY_NAME = "petals.log.dir";

    protected static final String DEFAULT_BASEDIR = "${" + PETALS_LOG_DIR_PROPERTY_NAME + "}";

    protected static final String FLOW_LOG_SUBDIR_PROPERTY_NAME = ".flows-subdir";

    protected static final String DEFAULT_FLOW_LOG_SUBDIR = "flow-monitoring";

    protected static final String LOGFILENAME_PROPERTY_NAME = ".logFilename";

    protected static final String DEFAULT_LOGFILENAME = "petals.log";

    protected static final String ERROR_LOGFILENAME_WITH_SEPARATOR = "The log filename cannot contain a file separator. Default value used: "
            + DEFAULT_LOGFILENAME;

    /**
     * The root directory (absolute) of log files.
     * 
     * Default value: {@value #DEFAULT_BASEDIR}
     */
    protected File basedir;

    /**
     * The root directory (absolute) of sub directory of flows.
     * 
     * If the value of the configuration property {@value #FLOW_LOG_SUBDIR_PROPERTY_NAME} is not an absolute path, the
     * flow sub dir is defined as sub-dir of {@link #basedir} .
     */
    protected final File flowsSubdir;

    protected String logfilename;

    /**
     * <p>
     * Create and configure the log handler from {@link LogManager} properties.
     * </p>
     * <p>
     * Note: If configuration error are detected, error are printed on the error output
     * </p>
     * 
     * @throws PropertiesException
     */
    public PetalsFileHandler() throws PropertiesException {

        super();

        // Initialize the basedir
        final String basedirValue;
        try {
            basedirValue = getStringProperty(BASEDIR_PROPERTY_NAME,
                    PropertiesHelper.resolveString(DEFAULT_BASEDIR, System.getProperties()));
        } catch (final PropertiesException e) {
            final String className = this.getClass().getName();
            System.err.println(className + ": the default value of " + className + BASEDIR_PROPERTY_NAME + " ("
                    + DEFAULT_BASEDIR + ") is invalid (bug): " + e.getMessage());
            // we shouldn't continue, this is a bug!
            throw e;
        }

        this.basedir = new File(basedirValue).getAbsoluteFile();

        if (!this.basedir.exists()) {
            if (!this.basedir.mkdirs()) {
                System.err.println(
                        this.getClass().getName() + ": Unable to create the base directory '" + this.basedir + "'");
            }
        }

        // Initialize the flow sub dir
        final String flowsSubDirValue = getStringProperty(FLOW_LOG_SUBDIR_PROPERTY_NAME, DEFAULT_FLOW_LOG_SUBDIR);

        final File flowsSubDirPath = new File(flowsSubDirValue);
        if (flowsSubDirPath.isAbsolute()) {
            this.flowsSubdir = flowsSubDirPath;
        } else {
            this.flowsSubdir = new File(this.basedir, flowsSubDirValue);
        }

        final String logfilename = getStringProperty(LOGFILENAME_PROPERTY_NAME, DEFAULT_LOGFILENAME);

        if (logfilename.contains(File.separator)) {
            System.err.println(ERROR_LOGFILENAME_WITH_SEPARATOR);
            this.logfilename = DEFAULT_LOGFILENAME;
        } else {
            this.logfilename = logfilename;
        }
    }

    @Override
    public synchronized void publish(final LogRecord record) {

        int levelValue = getLevel().intValue();
        if (record.getLevel().intValue() >= levelValue && levelValue != offValue) {

            this.printLog(record);
        }
    }

    protected void printLog(final LogRecord record) {

        final File logFile = this.getMonitLogFile(record);

        /*
         * FileOutputStream can generate java.io.IOException: Bad file descriptor if this FOS has been closed too early.
         * So, the publish method must keep the file locked until the publishing process is completed, in order to
         * close() the opening logging location file before another thread open new FOS base on a bad file descriptor
         * (because logging location file descriptor is already used).
         */
        if (logFile != null) {
            try {
                // open the File Output Stream in append mode (FOS) to the
                // log file
                final OutputStream fos = new FileOutputStream(logFile, true);

                // Set the FOS in order to allow the StreamHandler to write
                // in this FOS.
                super.setOutputStream(fos);

                // publish the log record.
                try {
                    super.publish(record);
                } finally {
                    super.close();
                }

            } catch (final FileNotFoundException e) {
                /*
                 * We can not throw an exception here, so report the exception to any registered ErrorManager.
                 */
                reportError(null, e, ErrorManager.OPEN_FAILURE);
            }
        }
    }

    /**
     * Get the log file where to write the given log record.
     * 
     * @param record
     *            Description of the log event
     * @return The {@link File} where the log record must be written.
     */
    private File getMonitLogFile(final LogRecord record) {

        final File flowInstanceSubdir = this.getFlowSubDir(record);
        final File logFileDir;
        if (flowInstanceSubdir == null) {
            logFileDir = this.basedir;
        } else {
            logFileDir = flowInstanceSubdir;
        }
        logFileDir.mkdirs();

        return new File(logFileDir, this.logfilename);
    }

    /**
     * <p>
     * Return the directory containing flow instance traces.
     * </p>
     * <p>
     * For MONIT level, the flow instance identifier is retrieved from a parameter of the log event. For other log
     * level, it is retrieved from the thread context.
     * </p>
     * 
     * @param record
     *            Description of the log event. If the level of the record is {@link Level#MONIT}, the attribute
     *            {@link PetalsExecutionContext#FLOW_INSTANCE_ID_PROPERTY_NAME} MUST be set as a non empty value into
     *            the parameters of the record.
     * @return The directory containing flow instance traces or <code>null</code> if no flow instance is in progress.
     */
    private File getFlowSubDir(final LogRecord record) {
        String flowInstanceId = null;
        if (Level.MONIT.equals(record.getLevel())) {
            final Object[] parameters = record.getParameters();
            if (parameters != null && parameters.length > 0) {
                final Object logDataObj = parameters[0];
                if (logDataObj instanceof LogData) {
                    flowInstanceId = (String) ((LogData) logDataObj).get(FlowLogData.FLOW_INSTANCE_ID_PROPERTY_NAME);
                    assert flowInstanceId != null : "A MONIT log event MUST always have an attribute 'flowInstanceId' set !!!";
                    assert !flowInstanceId.trim().isEmpty() : "A MONIT log event MUST always have a non-empty attribute 'flowInstanceId' !!!";
                } else {
                    reportError("A log trace with level MONIT must have a parameter as a FlowLogData instance", null,
                            ErrorManager.GENERIC_FAILURE);
                }
            } else {
                reportError("A log trace with level MONIT must have a parameter", null, ErrorManager.GENERIC_FAILURE);
            }
        }

        if (flowInstanceId == null) {
            flowInstanceId = PetalsExecutionContext.getFlowInstanceId();
        }

        if (flowInstanceId != null) {
            return new File(this.flowsSubdir, flowInstanceId);
        } else {
            return null;
        }
    }

    protected String getStringProperty(final String property, final String defaultValue) {
        final String className = getClass().getName();
        final String fullProperty = className + property;
        final String val = getAndResolve(fullProperty, defaultValue);
        if (val == null) {
            return defaultValue;
        }
        return val;
    }

    protected int getIntProperty(final String property, final int defaultValue) {
        final String className = getClass().getName();
        final String fullProperty = className + property;
        final String val = getAndResolve(fullProperty, defaultValue);
        if (val == null) {
            return defaultValue;
        }
        try {
            return Integer.parseInt(val);
        } catch (final Exception ex) {
            System.err.println(className + ": the value of " + fullProperty + " (" + val
                    + ") is not an int, default value (" + defaultValue + ") used: " + ex.getMessage());
            return defaultValue;
        }
    }

    protected String getAndResolve(final String fullProperty, final Object defaultValue) {
        final LogManager manager = LogManager.getLogManager();
        final String val = manager.getProperty(fullProperty);
        if (val == null || val.isEmpty()) {
            return null;
        }
        try {
            return PropertiesHelper.resolveString(val.trim(), System.getProperties());
        } catch (final PropertiesException e) {
            System.err.println(getClass().getName() + ": the value of " + fullProperty + "(" + val
                    + ") is invalid, default value (" + defaultValue + ") used: " + e.getMessage());
            return null;
        }
    }
}
