/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.log.formatter;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.LogRecord;

import org.junit.Before;
import org.junit.Test;
import org.ow2.petals.commons.log.FlowLogData;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.commons.log.TestLogData;
import org.ow2.petals.commons.log.TraceCode;
import org.ow2.petals.log.LogRecordHelper;

import com.ebmwebsourcing.easycommons.time.SimulatedTime;

public class LogDataFormatterTest extends AbstractLogDataFormatterTest {

    private static final String STARTING_DELIMITER = "{{";

    private static final String ENDING_DELIMITER = "}}";

    protected static final String INFORMATIONAL_MESS_PART = "This is a test message";

    protected SimulatedTime time = null;

    protected long utcTime = 0L;

    protected String formattedDate = null;

    @Before
    public void initSimulatedTime() {
        this.time = new SimulatedTime();
        // Stop the time (call me god )
        this.time.pause();
        this.utcTime = this.time.getUtcTime();
        this.formattedDate = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss,SSS 'GMT'Z",
                Locale.getDefault()).format(this.utcTime);
    }

    @Test
    public final void testLogDataAppenderPresentByDefault() {
        LogDataFormatter formatter = new LogDataFormatter();
        assertEquals(1, formatter.getLogDataAppenders().length);
        assertTrue(formatter.getLogDataAppenders()[0] instanceof BasicLogDataAppender);
    }

    @Test
    public final void testAddDataAppender() {
        LogDataFormatter formatter = new LogDataFormatter();
        BasicLogDataAppender blda1 = new BasicLogDataAppender();
        BasicLogDataAppender blda2 = new BasicLogDataAppender();
        formatter.addLogDataAppender(blda1);
        formatter.addLogDataAppender(blda2);
        assertEquals(3, formatter.getLogDataAppenders().length);
        assertTrue(formatter.getLogDataAppenders()[0] instanceof BasicLogDataAppender);
        assertEquals(blda1, formatter.getLogDataAppenders()[1]);
        assertEquals(blda2, formatter.getLogDataAppenders()[2]);
    }

    @Test
    public final void testFormattingWithoutExceptionNoRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        testFormatting(formatterUnderTest, null, null, null, "");
    }

    @Test
    public final void testFormattingWithExceptionNoRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        Throwable thrown = new RuntimeException("error...");
        testFormatting(formatterUnderTest, null, thrown, null, "");
    }

    @Test
    public final void testFormattingWithExceptionChainNoRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        Throwable cause2 = new RuntimeException("error cause2...");
        Throwable cause1 = new RuntimeException("error cause1...", cause2);
        Throwable thrown = new RuntimeException("error...", cause1);
        testFormatting(formatterUnderTest, null, thrown, null, "");
    }

    @Test
    public final void testFormattingWithoutExceptionUnhandledRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        testFormatting(formatterUnderTest, new Object[] { 1, "fakeparam", 2.3 }, null, null, "");
    }

    @Test
    public final void testFormattingWithoutExceptionLogDataAsRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter(STARTING_DELIMITER, ENDING_DELIMITER);
        testFormatting(formatterUnderTest, new Object[] { createTestLogData() }, null, null,
                " " + STARTING_DELIMITER + "key1 = 'value1', key2 = 'value2', key3 = 'value3'"
                + ENDING_DELIMITER);
    }

    @Test
    public final void testFormattingWithoutExceptionEmptyLogDataAsRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        testFormatting(formatterUnderTest, new Object[] { new TestLogData() }, null, null, "");
    }

    @Test
    public final void testFormattingWithExceptionLogDataAsRecordParam() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter(STARTING_DELIMITER, ENDING_DELIMITER);
        Throwable thrown = new RuntimeException("error...");
        testFormatting(formatterUnderTest, new Object[] { createTestLogData() }, thrown, null,
                " " + STARTING_DELIMITER + "key1 = 'value1', key2 = 'value2', key3 = 'value3'"
                + ENDING_DELIMITER);
    }

    @Test
    public final void testFormatMessageWithNullAsParametersValue() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        testFormatting(formatterUnderTest, null, null, null, "");
    }

    @Test
    public final void testFormatMessageWithNullAsFirstValueInParameters() throws Exception {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        testFormatting(formatterUnderTest, new String[] { null }, null, null, "");
    }

    @Test
    public void testFormatMessageWithPrefix() {
        LogDataFormatter formatterUnderTest = new LogDataFormatter();
        final String expectedPrefix = "petals-node1";
        PetalsExecutionContext.putContainerName(expectedPrefix);
        try {
            testFormatting(formatterUnderTest, null, null, expectedPrefix, "");
        } finally {
            PetalsExecutionContext.clearAll();
        }
    }
    
    protected void testFormatting(LogDataFormatter formatterUnderTest, Object[] logDataParameters,
            Throwable thrown, String expectedPrefix, String expectedExtraParametersMessPart) {

        final Level level = Level.INFO;
        LogRecord logRecord = thrown != null ? LogRecordHelper.newLogRecordWithException(level,
                INFORMATIONAL_MESS_PART, this.utcTime, thrown) : LogRecordHelper.newLogRecord(
                level, INFORMATIONAL_MESS_PART, this.utcTime);
        if (logDataParameters != null) {
            logRecord.setParameters(logDataParameters);
        }
        String expectedLogMessage = (expectedPrefix != null ? expectedPrefix : "") + " "
                + this.formattedDate + " " + level + " [" + logRecord.getLoggerName() + "] : "
                + INFORMATIONAL_MESS_PART
                + expectedExtraParametersMessPart + "\n"
                + (thrown != null ? generateExpectedExceptionMessagePart(thrown) : "");

        String myLog = formatterUnderTest.format(logRecord);

        assertEquals(expectedLogMessage, myLog);

    }
    
    protected String generateExpectedExceptionMessagePart(Throwable throwable) {
        final StringWriter sw = new StringWriter();
        final PrintWriter pw = new PrintWriter(sw);
        throwable.printStackTrace(pw);
        pw.println();
        pw.close();
        return sw.toString();
    }
    

    class TestLogDataWithExtraParam extends FlowLogData {
        
        private static final long serialVersionUID = -3323328847345282397L;

        public static final String HIDDEN_EXTRA_PARAM = "hiddenExtraParam";

        public TestLogDataWithExtraParam() {
            super(TraceCode.PROVIDE_FLOW_STEP_END);
            putData(HIDDEN_EXTRA_PARAM, "paramValue");
        }

    }

    /**
     * Check that the formatter does not print an extra parameter that should
     * not be printed.
     */
    @Test
    public final void testFormatMessageWithExtraParameterToNotPrint() throws Exception {

        // Generate the log conf file
        final File logConfFile = this.generateLogConfFile(new ConfPropertiesSetter() {

            @Override
            public void customizeProperties(final Properties confProperties) throws Exception {
                confProperties.setProperty(LogDataFormatter.class.getName() + ".appender",
                        TestLogDataAppenderHideningExtraParam.class.getName());
            }

        });

        final LogManager manager = LogManager.getLogManager();
        final InputStream isConfLog = new FileInputStream(logConfFile);
        try {
            manager.readConfiguration(isConfLog);
        } finally {
            isConfLog.close();
        }

        final LogDataFormatter formatter = new LogDataFormatter();
        final String msg = "This is a test message";
        final LogRecord logRecord = LogRecordHelper.newLogRecord(Level.INFO, msg, this.utcTime);
        logRecord.setParameters(new Object[] { new TestLogDataWithExtraParam() });
        final String myLog = formatter.format(logRecord);
        assertFalse("The hidden extra param is printed",
                myLog.contains(TestLogDataWithExtraParam.HIDDEN_EXTRA_PARAM));
        assertTrue("The end of the message is not the expected one",
                myLog.endsWith("traceCode = 'provideFlowStepEnd'\n"));
    }
}
